import numpy as np
import pytest

from anemoi.graphs.normalizer import NormalizerMixin


@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-std"])
def test_normalizer(norm: str):
    """Test NormalizerMixin normalize method."""

    class Normalizer(NormalizerMixin):
        def __init__(self, norm):
            self.norm = norm

        def __call__(self, data):
            return self.normalize(data)

    normalizer = Normalizer(norm=norm)
    data = np.random.rand(10, 5)
    normalized_data = normalizer(data)
    assert isinstance(normalized_data, np.ndarray)
    assert normalized_data.shape == data.shape


@pytest.mark.parametrize("norm", ["l3", "invalid"])
def test_normalizer_wrong_norm(norm: str):
    """Test NormalizerMixin normalize method."""

    class Normalizer(NormalizerMixin):
        def __init__(self, norm: str):
            self.norm = norm

        def __call__(self, data):
            return self.normalize(data)

    with pytest.raises(ValueError):
        normalizer = Normalizer(norm=norm)
        data = np.random.rand(10, 5)
        normalizer(data)


def test_normalizer_wrong_inheritance():
    """Test NormalizerMixin normalize method."""

    class Normalizer(NormalizerMixin):
        def __init__(self, attr):
            self.attr = attr

        def __call__(self, data):
            return self.normalize(data)

    with pytest.raises(AttributeError):
        normalizer = Normalizer(attr="attr_name")
        data = np.random.rand(10, 5)
        normalizer(data)



import numpy as np
import pytest
import torch
import yaml
from torch_geometric.data import HeteroData

lats = [-0.15, 0, 0.15]
lons = [0, 0.25, 0.5, 0.75]


class MockZarrDataset:
    """Mock Zarr dataset with latitudes and longitudes attributes."""

    def __init__(self, latitudes, longitudes):
        self.latitudes = latitudes
        self.longitudes = longitudes
        self.num_nodes = len(latitudes)


@pytest.fixture
def mock_zarr_dataset() -> MockZarrDataset:
    """Mock zarr dataset with nodes."""
    coords = 2 * torch.pi * np.array([[lat, lon] for lat in lats for lon in lons])
    return MockZarrDataset(latitudes=coords[:, 0], longitudes=coords[:, 1])


@pytest.fixture
def mock_grids_path(tmp_path) -> tuple[str, int]:
    """Mock grid_definition_path with files for 3 resolutions."""
    num_nodes = len(lats) * len(lons)
    for resolution in ["o16", "o48", "5km5"]:
        file_path = tmp_path / f"grid-{resolution}.npz"
        np.savez(file_path, latitudes=np.random.rand(num_nodes), longitudes=np.random.rand(num_nodes))
    return str(tmp_path), num_nodes


@pytest.fixture
def graph_with_nodes() -> HeteroData:
    """Graph with 12 nodes over the globe, stored in \"test_nodes\"."""
    coords = np.array([[lat, lon] for lat in lats for lon in lons])
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor(coords)
    return graph


@pytest.fixture
def graph_nodes_and_edges() -> HeteroData:
    """Graph with 1 set of nodes and edges."""
    coords = np.array([[lat, lon] for lat in lats for lon in lons])
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor(coords)
    graph[("test_nodes", "to", "test_nodes")].edge_index = torch.tensor([[0, 1], [1, 2], [2, 3], [3, 0]])
    return graph


@pytest.fixture
def config_file(tmp_path) -> tuple[str, str]:
    """Mock grid_definition_path with files for 3 resolutions."""
    cfg = {
        "nodes": {
            "test_nodes": {
                "node_builder": {
                    "_target_": "anemoi.graphs.nodes.NPZFileNodes",
                    "grid_definition_path": str(tmp_path),
                    "resolution": "o16",
                },
            },
        },
        "edges": [
            {
                "source_name": "test_nodes",
                "target_name": "test_nodes",
                "edge_builder": {
                    "_target_": "anemoi.graphs.edges.KNNEdges",
                    "num_nearest_neighbours": 3,
                },
                "attributes": {
                    "dist_norm": {"_target_": "anemoi.graphs.edges.attributes.EdgeLength"},
                    "edge_dirs": {"_target_": "anemoi.graphs.edges.attributes.EdgeDirection"},
                },
            },
        ],
    }
    file_name = "config.yaml"

    with (tmp_path / file_name).open("w") as file:
        yaml.dump(cfg, file)

    return tmp_path, file_name



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from pathlib import Path

import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.create import GraphCreator


class TestGraphCreator:

    @pytest.mark.parametrize("name", ["graph.pt", None])
    def test_generate_graph(self, config_file: tuple[Path, str], mock_grids_path: tuple[str, int], name: str):
        """Test GraphCreator workflow."""
        tmp_path, config_name = config_file
        graph_path = tmp_path / name if isinstance(name, str) else None
        config_path = tmp_path / config_name

        graph = GraphCreator(config=config_path).create(save_path=graph_path)

        assert isinstance(graph, HeteroData)
        assert "test_nodes" in graph.node_types
        assert ("test_nodes", "to", "test_nodes") in graph.edge_types

        for nodes in graph.node_stores:
            for node_attr in nodes.node_attrs():
                assert isinstance(nodes[node_attr], torch.Tensor)
                assert nodes[node_attr].dtype in [torch.int32, torch.float32]

        for edges in graph.edge_stores:
            for edge_attr in edges.edge_attrs():
                assert isinstance(edges[edge_attr], torch.Tensor)
                assert edges[edge_attr].dtype in [torch.int32, torch.float32]

        for nodes in graph.node_stores:
            for node_attr in nodes.node_attrs():
                assert not node_attr.startswith("_")
        for edges in graph.edge_stores:
            for edge_attr in edges.edge_attrs():
                assert not edge_attr.startswith("_")

        if graph_path is not None:
            assert graph_path.exists()
            graph_saved = torch.load(graph_path)
            assert graph.node_types == graph_saved.node_types
            assert graph.edge_types == graph_saved.edge_types



import logging
from pathlib import Path
from typing import Optional
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from torch_geometric.data import HeteroData

from anemoi.graphs.plotting.prepare import compute_isolated_nodes
from anemoi.graphs.plotting.prepare import compute_node_adjacencies
from anemoi.graphs.plotting.prepare import edge_list
from anemoi.graphs.plotting.prepare import node_list

annotations_style = {"text": "", "showarrow": False, "xref": "paper", "yref": "paper", "x": 0.005, "y": -0.002}
plotly_axis_config = {"showgrid": False, "zeroline": False, "showticklabels": False}

LOGGER = logging.getLogger(__name__)


def plot_interactive_subgraph(
    graph: HeteroData,
    edges_to_plot: tuple[str, str, str],
    out_file: Optional[Union[str, Path]] = None,
) -> None:
    """Plots a bipartite graph (bi-graph).

    This methods plots the bipartite graph passed in an interactive window (using Ploty).

    Parameters
    ----------
    graph : dict
        The graph to plot.
    edges_to_plot : tuple[str, str]
        Names of the edges to plot.
    out_file : str | Path, optional
        Name of the file to save the plot. Default is None.
    """
    source_name, _, target_name = edges_to_plot
    edge_x, edge_y = edge_list(graph, source_nodes_name=source_name, target_nodes_name=target_name)
    assert source_name in graph.node_types, f"edges_to_plot ({source_name}) should be in the graph"
    assert target_name in graph.node_types, f"edges_to_plot ({target_name}) should be in the graph"
    lats_source_nodes, lons_source_nodes = node_list(graph, source_name)
    lats_target_nodes, lons_target_nodes = node_list(graph, target_name)

    # Compute node adjacencies
    node_adjacencies = compute_node_adjacencies(graph, source_name, target_name)
    node_text = [f"# of connections: {x}" for x in node_adjacencies]

    edge_trace = go.Scattergeo(
        lat=edge_x,
        lon=edge_y,
        line={"width": 0.5, "color": "#888"},
        hoverinfo="none",
        mode="lines",
        name="Connections",
    )

    source_node_trace = go.Scattergeo(
        lat=lats_source_nodes,
        lon=lons_source_nodes,
        mode="markers",
        hoverinfo="text",
        name=source_name,
        marker={
            "showscale": False,
            "color": "red",
            "size": 2,
            "line_width": 2,
        },
    )

    target_node_trace = go.Scattergeo(
        lat=lats_target_nodes,
        lon=lons_target_nodes,
        mode="markers",
        hoverinfo="text",
        name=target_name,
        text=node_text,
        marker={
            "showscale": True,
            "colorscale": "YlGnBu",
            "reversescale": True,
            "color": list(node_adjacencies),
            "size": 10,
            "colorbar": {"thickness": 15, "title": "Node Connections", "xanchor": "left", "titleside": "right"},
            "line_width": 2,
        },
    )
    layout = go.Layout(
        title="<br>" + f"Graph {source_name} --> {target_name}",
        titlefont_size=16,
        showlegend=True,
        hovermode="closest",
        margin={"b": 20, "l": 5, "r": 5, "t": 40},
        annotations=[annotations_style],
        legend={"x": 0, "y": 1},
        xaxis=plotly_axis_config,
        yaxis=plotly_axis_config,
    )
    fig = go.Figure(data=[edge_trace, source_node_trace, target_node_trace], layout=layout)
    fig.update_geos(fitbounds="locations")

    if out_file is not None:
        fig.write_html(out_file)
    else:
        fig.show()


def plot_isolated_nodes(graph: HeteroData, out_file: Optional[Union[str, Path]] = None) -> None:
    """Plot isolated nodes.

    This method creates an interactive visualization of the isolated nodes in the graph.

    Parameters
    ----------
    graph : AnemoiGraph
        The graph to plot.
    out_file : str | Path, optional
        Name of the file to save the plot. Default is None.
    """
    isolated_nodes = compute_isolated_nodes(graph)

    if len(isolated_nodes) == 0:
        LOGGER.warning("No isolated nodes found.")
        return

    colorbar = plt.cm.rainbow(np.linspace(0, 1, len(isolated_nodes)))
    nodes = []
    for name, (lat, lon) in isolated_nodes.items():
        nodes.append(
            go.Scattergeo(
                lat=lat,
                lon=lon,
                mode="markers",
                hoverinfo="text",
                name=name,
                marker={"showscale": False, "color": colorbar[len(nodes)], "size": 10},
            ),
        )

    layout = go.Layout(
        title="<br>Orphan nodes",
        titlefont_size=16,
        showlegend=True,
        hovermode="closest",
        margin={"b": 20, "l": 5, "r": 5, "t": 40},
        annotations=[annotations_style],
        legend={"x": 0, "y": 1},
        xaxis=plotly_axis_config,
        yaxis=plotly_axis_config,
    )
    fig = go.Figure(data=nodes, layout=layout)
    fig.update_geos(fitbounds="locations")

    if out_file is not None:
        fig.write_html(out_file)
    else:
        fig.show()


def plot_interactive_nodes(graph: HeteroData, nodes_name: str, out_file: Optional[str] = None) -> None:
    """Plot nodes.

    This method creates an interactive visualization of a set of nodes.

    Parameters
    ----------
    graph : HeteroData
        Graph.
    nodes_name : str
        Name of the nodes to plot.
    out_file : str, optional
        Name of the file to save the plot. Default is None.
    """
    node_latitudes, node_longitudes = node_list(graph, nodes_name)
    node_attrs = graph[nodes_name].node_attrs()
    # Remove x to avoid plotting the coordinates as an attribute
    node_attrs.remove("x")

    if len(node_attrs) == 0:
        LOGGER.warning(f"No node attributes found for {nodes_name} nodes.")
        return

    node_traces = {}
    for node_attr in node_attrs:
        node_attr_values = graph[nodes_name][node_attr].float().numpy()

        # Skip multi-dimensional attributes. Supported only: (N, 1) or (N,) tensors
        if node_attr_values.ndim > 1 and node_attr_values.shape[1] > 1:
            continue

        node_traces[node_attr] = go.Scattergeo(
            lat=node_latitudes,
            lon=node_longitudes,
            name=" ".join(node_attr.split("_")).capitalize(),
            mode="markers",
            hoverinfo="text",
            marker={
                "color": node_attr_values.squeeze().tolist(),
                "showscale": True,
                "colorscale": "RdBu",
                "colorbar": {"thickness": 15, "title": node_attr, "xanchor": "left"},
                "size": 5,
            },
            visible=False,
        )

    # Create and add slider
    slider_steps = []
    for i, node_attr in enumerate(node_traces.keys()):
        step = dict(
            label=f"Node attribute: {node_attr}",
            method="update",
            args=[{"visible": [False] * len(node_traces)}],
        )
        step["args"][0]["visible"][i] = True  # Toggle i'th trace to "visible"
        slider_steps.append(step)

    fig = go.Figure(
        data=list(node_traces.values()),
        layout=go.Layout(
            title=f"<br>Map of {nodes_name} nodes",
            sliders=[
                dict(active=0, currentvalue={"visible": False}, len=0.4, x=0.5, xanchor="center", steps=slider_steps)
            ],
            titlefont_size=16,
            showlegend=False,
            hovermode="closest",
            margin={"b": 20, "l": 5, "r": 5, "t": 40},
            annotations=[annotations_style],
            xaxis=plotly_axis_config,
            yaxis=plotly_axis_config,
        ),
    )
    fig.data[0].visible = True

    if out_file is not None:
        fig.write_html(out_file)
    else:
        fig.show()



from __future__ import annotations

import numpy as np
from scipy.spatial.transform import Rotation

from anemoi.graphs.generate.transforms import direction_vec
from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian


def get_rotation_from_unit_vecs(points: np.ndarray, reference: np.ndarray) -> Rotation:
    """Compute rotation matrix of a set of points with respect to a reference vector.

    Parameters
    ----------
    points : np.ndarray of shape (num_points, 3)
        The points to compute the direction vector.
    reference : np.ndarray of shape (3, )
        The reference vector.

    Returns
    -------
    Rotation
        The rotation matrix that aligns the points with the reference vector.
    """
    assert points.shape[1] == 3, "Points must be in 3D"
    v_unit = direction_vec(points, reference)
    theta = np.arccos(np.dot(points, reference))
    return Rotation.from_rotvec(np.transpose(v_unit * theta))


def compute_directions(loc1: np.ndarray, loc2: np.ndarray, pole_vec: np.ndarray | None = None) -> np.ndarray:
    """Compute the direction of the edge joining the nodes considered.

    Parameters
    ----------
    loc1 : np.ndarray of shape (2, num_points)
        Location of the head nodes.
    loc2 : np.ndarray of shape (2, num_points)
        Location of the tail nodes.
    pole_vec : np.ndarray, optional
        The pole vector to rotate the points to. Defaults to the north pole.

    Returns
    -------
    np.ndarray of shape (3, num_points)
        The direction of the edge after rotating the north pole.
    """
    if pole_vec is None:
        pole_vec = np.array([0, 0, 1])

    # all will be rotated relative to destination node
    loc1_xyz = latlon_rad_to_cartesian(loc1, 1.0)
    loc2_xyz = latlon_rad_to_cartesian(loc2, 1.0)
    r = get_rotation_from_unit_vecs(loc2_xyz, pole_vec)
    direction = direction_vec(r.apply(loc1_xyz), pole_vec)
    return direction / np.sqrt(np.power(direction, 2).sum(axis=0))


def directional_edge_features(
    loc1: np.ndarray, loc2: np.ndarray, relative_to_rotated_target: bool = True
) -> np.ndarray:
    """Compute features of the edge joining the nodes considered.

    It computes the direction of the edge after rotating the north pole.

    Parameters
    ----------
    loc1 : np.ndarray of shpae (2, num_points)
        Location of the head node.
    loc2 : np.ndarray of shape (2, num_points)
        Location of the tail node.
    relative_to_rotated_target : bool, optional
        Whether to rotate the north pole to the target node. Defaults to True.

    Returns
    -------
    np.ndarray of shape of (2, num_points)
        Direction of the edge after rotation the north pole.
    """
    if relative_to_rotated_target:
        rotation = compute_directions(loc1, loc2)
        assert np.allclose(rotation[2], 0), "Rotation should be aligned with the north pole"
        return rotation[:2]

    return loc2 - loc1



import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes import TriNodes
from anemoi.graphs.nodes.builder import BaseNodeBuilder


@pytest.mark.parametrize("resolution", [0, 2])
def test_init(resolution: list[int]):
    """Test TrirefinedIcosahedralNodes initialization."""

    node_builder = TriNodes(resolution, "test_nodes")
    assert isinstance(node_builder, BaseNodeBuilder)
    assert isinstance(node_builder, TriNodes)


def test_get_coordinates():
    """Test get_coordinates method."""
    node_builder = TriNodes(2, "test_nodes")
    coords = node_builder.get_coordinates()
    assert isinstance(coords, torch.Tensor)
    assert coords.shape == (162, 2)


def test_update_graph():
    """Test update_graph method."""
    node_builder = TriNodes(1, "test_nodes")
    graph = HeteroData()
    graph = node_builder.update_graph(graph, {})
    assert "_resolutions" in graph["test_nodes"]
    assert "_nx_graph" in graph["test_nodes"]
    assert "_node_ordering" in graph["test_nodes"]
    assert len(graph["test_nodes"]["_node_ordering"]) == graph["test_nodes"].num_nodes



from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod

import numpy as np
import torch
from scipy.spatial import SphericalVoronoi
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian
from anemoi.graphs.normalizer import NormalizerMixin

LOGGER = logging.getLogger(__name__)


class BaseWeights(ABC, NormalizerMixin):
    """Base class for the weights of the nodes."""

    def __init__(self, norm: str | None = None) -> None:
        self.norm = norm

    @abstractmethod
    def get_raw_values(self, nodes: NodeStorage, *args, **kwargs): ...

    def post_process(self, values: np.ndarray) -> torch.Tensor:
        """Post-process the values."""
        if values.ndim == 1:
            values = values[:, np.newaxis]

        norm_values = self.normalize(values)

        return torch.tensor(norm_values, dtype=torch.float32)

    def compute(self, graph: HeteroData, nodes_name: str, *args, **kwargs) -> torch.Tensor:
        """Get the node weights.

        Parameters
        ----------
        graph : HeteroData
            Graph.
        nodes_name : str
            Name of the nodes.

        Returns
        -------
        torch.Tensor
            Weights associated to the nodes.
        """
        nodes = graph[nodes_name]
        weights = self.get_raw_values(nodes, *args, **kwargs)
        return self.post_process(weights)


class UniformWeights(BaseWeights):
    """Implements a uniform weight for the nodes.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def get_raw_values(self, nodes: NodeStorage, *args, **kwargs) -> np.ndarray:
        """Compute the weights.

        Parameters
        ----------
        nodes : NodeStorage
            Nodes of the graph.

        Returns
        -------
        np.ndarray
            Weights.
        """
        return np.ones(nodes.num_nodes)


class AreaWeights(BaseWeights):
    """Implements the area of the nodes as the weights.

    Attributes
    ----------
    norm : str
        Normalization of the weights.
    radius : float
        Radius of the sphere.
    centre : np.ndarray
        Centre of the sphere.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def __init__(self, norm: str | None = None, radius: float = 1.0, centre: np.ndarray = np.array([0, 0, 0])) -> None:
        super().__init__(norm)
        self.radius = radius
        self.centre = centre

    def get_raw_values(self, nodes: NodeStorage, *args, **kwargs) -> np.ndarray:
        """Compute the area associated to each node.

        It uses Voronoi diagrams to compute the area of each node.

        Parameters
        ----------
        nodes : NodeStorage
            Nodes of the graph.

        Returns
        -------
        np.ndarray
            Weights.
        """
        latitudes, longitudes = nodes.x[:, 0], nodes.x[:, 1]
        points = latlon_rad_to_cartesian((np.asarray(latitudes), np.asarray(longitudes)))
        sv = SphericalVoronoi(points, self.radius, self.centre)
        area_weights = sv.calculate_areas()
        LOGGER.debug(
            "There are %d of weights, which (unscaled) add up a total weight of %.2f.",
            len(area_weights),
            np.array(area_weights).sum(),
        )
        return area_weights



import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import AreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights
from anemoi.graphs.nodes.builder import NPZFileNodes


@pytest.mark.parametrize("resolution", ["o16", "o48", "5km5"])
def test_init(mock_grids_path: tuple[str, int], resolution: str):
    """Test NPZFileNodes initialization."""
    grid_definition_path, _ = mock_grids_path
    node_builder = NPZFileNodes(resolution, grid_definition_path, "test_nodes")
    assert isinstance(node_builder, NPZFileNodes)


@pytest.mark.parametrize("resolution", ["o17", 13, "ajsnb", None])
def test_fail_init_wrong_resolution(mock_grids_path: tuple[str, int], resolution: str):
    """Test NPZFileNodes initialization with invalid resolution."""
    grid_definition_path, _ = mock_grids_path
    with pytest.raises(FileNotFoundError):
        NPZFileNodes(resolution, grid_definition_path, "test_nodes")


def test_fail_init_wrong_path():
    """Test NPZFileNodes initialization with invalid path."""
    with pytest.raises(FileNotFoundError):
        NPZFileNodes("o16", "invalid_path", "test_nodes")


@pytest.mark.parametrize("resolution", ["o16", "o48", "5km5"])
def test_register_nodes(mock_grids_path: str, resolution: str):
    """Test NPZFileNodes register correctly the nodes."""
    graph = HeteroData()
    grid_definition_path, num_nodes = mock_grids_path
    node_builder = NPZFileNodes(resolution, grid_definition_path, "test_nodes")

    graph = node_builder.register_nodes(graph)

    assert graph["test_nodes"].x is not None
    assert isinstance(graph["test_nodes"].x, torch.Tensor)
    assert graph["test_nodes"].x.shape == (num_nodes, 2)
    assert graph["test_nodes"].node_type == "NPZFileNodes"


@pytest.mark.parametrize("attr_class", [UniformWeights, AreaWeights])
def test_register_attributes(graph_with_nodes: HeteroData, mock_grids_path: tuple[str, int], attr_class):
    """Test NPZFileNodes register correctly the weights."""
    grid_definition_path, _ = mock_grids_path
    node_builder = NPZFileNodes("o16", grid_definition_path, "test_nodes")
    config = {"test_attr": {"_target_": f"anemoi.graphs.nodes.attributes.{attr_class.__name__}"}}

    graph = node_builder.register_attributes(graph_with_nodes, config)

    assert graph["test_nodes"]["test_attr"] is not None
    assert isinstance(graph["test_nodes"]["test_attr"], torch.Tensor)
    assert graph["test_nodes"]["test_attr"].shape[0] == graph["test_nodes"].x.shape[0]



import argparse

from anemoi.graphs.descriptor import GraphDescriptor
from anemoi.graphs.inspector import GraphInspector

from . import Command


class Inspect(Command):
    """Inspect a graph."""

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument(
            "--show_attribute_distributions",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Hide distribution plots of edge/node attributes.",
        )
        command_parser.add_argument(
            "--show_nodes",
            action=argparse.BooleanOptionalAction,
            default=False,
            help="Show the nodes of the graph.",
        )
        command_parser.add_argument(
            "--description",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Hide the description of the graph.",
        )
        command_parser.add_argument("path", help="Path to the graph (a .PT file).")
        command_parser.add_argument("output_path", help="Path to store the inspection results.")

    def run(self, args):
        kwargs = vars(args)

        if kwargs.get("description", False):
            GraphDescriptor(kwargs["path"]).describe()

        inspector = GraphInspector(**kwargs)
        inspector.inspect()


command = Inspect



import math
from itertools import chain
from pathlib import Path
from typing import Optional
from typing import Union

import torch
from anemoi.utils.humanize import bytes
from anemoi.utils.text import table


class GraphDescriptor:
    """Class for descripting the graph."""

    def __init__(self, path: Union[str, Path], **kwargs):
        self.path = path
        self.graph = torch.load(self.path)

    @property
    def total_size(self):
        """Total size of the tensors in the graph (in bytes)."""
        total_size = 0

        for store in chain(self.graph.node_stores, self.graph.edge_stores):
            for value in store.values():
                if isinstance(value, torch.Tensor):
                    total_size += value.numel() * value.element_size()

        return total_size

    def get_node_summary(self) -> list[list]:
        """Summary of the nodes in the graph.

        Returns
        -------
        list[list]
            Returns a list for each subgraph with the following information:
            - Node name.
            - Number of nodes.
            - List of attribute names.
            - Total dimension of the attributes.
            - Min. latitude.
            - Max. latitude.
            - Min. longitude.
            - Max. longitude.
        """
        node_summary = []
        for name, nodes in self.graph.node_items():
            attributes = nodes.node_attrs()
            attributes.remove("x")

            node_summary.append(
                [
                    name,
                    nodes.num_nodes,
                    ", ".join(attributes),
                    sum(nodes[attr].shape[1] for attr in attributes if isinstance(nodes[attr], torch.Tensor)),
                    nodes.x[:, 0].min().item() / 2 / math.pi * 360,
                    nodes.x[:, 0].max().item() / 2 / math.pi * 360,
                    nodes.x[:, 1].min().item() / 2 / math.pi * 360,
                    nodes.x[:, 1].max().item() / 2 / math.pi * 360,
                ]
            )
        return node_summary

    def get_edge_summary(self) -> list[list]:
        """Summary of the edges in the graph.

        Returns
        -------
        list[list]
            Returns a list for each subgraph with the following information:
            - Source node name.
            - Destination node name.
            - Number of edges.
            - Number of isolated source nodes.
            - Number of isolated target nodes.
            - Total dimension of the attributes.
            - List of attribute names.
        """
        edge_summary = []
        for (src_name, _, dst_name), edges in self.graph.edge_items():
            attributes = edges.edge_attrs()
            attributes.remove("edge_index")

            edge_summary.append(
                [
                    src_name,
                    dst_name,
                    edges.num_edges,
                    self.graph[src_name].num_nodes - len(torch.unique(edges.edge_index[0])),
                    self.graph[dst_name].num_nodes - len(torch.unique(edges.edge_index[1])),
                    sum(edges[attr].shape[1] for attr in attributes),
                    ", ".join([f"{attr}({edges[attr].shape[1]}D)" for attr in attributes]),
                ]
            )
        return edge_summary

    def get_node_attribute_table(self) -> list[list]:
        node_attributes = []
        for node_name, node_store in self.graph.node_items():
            node_attr_names = node_store.node_attrs()
            node_attr_names.remove("x")  # Remove the coordinates from statistics table
            for node_attr_name in node_attr_names:
                node_attributes.append(
                    [
                        "Node",
                        node_name,
                        node_attr_name,
                        node_store[node_attr_name].dtype,
                        node_store[node_attr_name].float().min().item(),
                        node_store[node_attr_name].float().mean().item(),
                        node_store[node_attr_name].float().max().item(),
                        node_store[node_attr_name].float().std().item(),
                    ]
                )
        return node_attributes

    def get_edge_attribute_table(self) -> list[list]:
        edge_attributes = []
        for (source_name, _, target_name), edge_store in self.graph.edge_items():
            edge_attr_names = edge_store.edge_attrs()
            edge_attr_names.remove("edge_index")  # Remove the edge index from statistics table
            for edge_attr_name in edge_attr_names:
                edge_attributes.append(
                    [
                        "Edge",
                        f"{source_name}-->{target_name}",
                        edge_attr_name,
                        edge_store[edge_attr_name].dtype,
                        edge_store[edge_attr_name].float().min().item(),
                        edge_store[edge_attr_name].float().mean().item(),
                        edge_store[edge_attr_name].float().max().item(),
                        edge_store[edge_attr_name].float().std().item(),
                    ]
                )

        return edge_attributes

    def get_attribute_table(self) -> list[list]:
        """Get a table with the attributes of the graph."""
        attribute_table = []
        attribute_table.extend(self.get_node_attribute_table())
        attribute_table.extend(self.get_edge_attribute_table())
        return attribute_table

    def describe(self, show_attribute_distributions: Optional[bool] = True) -> None:
        """Describe the graph."""
        print()
        print(f"📦 Path       : {self.path}")
        print(f"💽 Size       : {bytes(self.total_size)} ({self.total_size})")
        print()
        print("🪩  Nodes summary")
        print()
        print(
            table(
                self.get_node_summary(),
                header=[
                    "Nodes name",
                    "Num. nodes",
                    "Attributes",
                    "Attribute dim",
                    "Min. latitude",
                    "Max. latitude",
                    "Min. longitude",
                    "Max. longitude",
                ],
                align=["<", ">", ">", ">", ">", ">", ">", ">"],
                margin=3,
            )
        )
        print()
        print()
        print("🌐  Edges summary")
        print()
        print(
            table(
                self.get_edge_summary(),
                header=[
                    "Source",
                    "Target",
                    "Num. edges",
                    "Isolated Source",
                    "Isolated Target",
                    "Attribute dim",
                    "Attributes",
                ],
                align=["<", "<", ">", ">", ">", ">", ">"],
                margin=3,
            )
        )
        print()
        if show_attribute_distributions:
            print()
            print("📊 Attribute distributions")
            print()
            print(
                table(
                    self.get_attribute_table(),
                    header=[
                        "Type",
                        "Source",
                        "Name",
                        "Dtype",
                        "Min.",
                        "Mean",
                        "Max.",
                        "Std. dev.",
                    ],
                    align=["<", "<", ">", ">", ">", ">", ">", ">"],
                    margin=3,
                )
            )
            print()
        print("🔋 Graph ready.")
        print()



import argparse
import logging
from pathlib import Path

from anemoi.graphs.create import GraphCreator
from anemoi.graphs.descriptor import GraphDescriptor

from . import Command

LOGGER = logging.getLogger(__name__)


class Create(Command):
    """Create a graph."""

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument(
            "--overwrite",
            action="store_true",
            help="Overwrite existing files. This will delete the target graph if it already exists.",
        )
        command_parser.add_argument(
            "--description",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Show the description of the graph.",
        )
        command_parser.add_argument(
            "config", type=Path, help="Configuration yaml file path defining the recipe to create the graph."
        )
        command_parser.add_argument("save_path", type=Path, help="Path to store the created graph.")

    def run(self, args):
        graph_creator = GraphCreator(config=args.config)
        graph_creator.create(save_path=args.save_path, overwrite=args.overwrite)

        if args.description:
            if args.save_path.exists():
                GraphDescriptor(args.save_path).describe()
            else:
                print("Graph description is not shown if the graph is not saved.")


command = Create



import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import AreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights


@pytest.mark.parametrize("norm", [None, "l1", "l2", "unit-max", "unit-std"])
def test_uniform_weights(graph_with_nodes: HeteroData, norm: str):
    """Test attribute builder for UniformWeights."""
    node_attr_builder = UniformWeights(norm=norm)
    weights = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert weights is not None
    assert isinstance(weights, torch.Tensor)
    assert weights.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]


@pytest.mark.parametrize("norm", ["l3", "invalide"])
def test_uniform_weights_fail(graph_with_nodes: HeteroData, norm: str):
    """Test attribute builder for UniformWeights with invalid norm."""
    with pytest.raises(ValueError):
        node_attr_builder = UniformWeights(norm=norm)
        node_attr_builder.compute(graph_with_nodes, "test_nodes")


def test_area_weights(graph_with_nodes: HeteroData):
    """Test attribute builder for AreaWeights."""
    node_attr_builder = AreaWeights()
    weights = node_attr_builder.compute(graph_with_nodes, "test_nodes")

    assert weights is not None
    assert isinstance(weights, torch.Tensor)
    assert weights.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]


@pytest.mark.parametrize("radius", [-1.0, "hello", None])
def test_area_weights_fail(graph_with_nodes: HeteroData, radius: float):
    """Test attribute builder for AreaWeights with invalid radius."""
    with pytest.raises(ValueError):
        node_attr_builder = AreaWeights(radius=radius)
        node_attr_builder.compute(graph_with_nodes, "test_nodes")



from .builder import HexNodes
from .builder import NPZFileNodes
from .builder import TriNodes
from .builder import ZarrDatasetNodes

__all__ = ["ZarrDatasetNodes", "NPZFileNodes", "TriNodes", "HexNodes"]



import pytest
from torch_geometric.data import HeteroData

from anemoi.graphs.edges import MultiScaleEdges
from anemoi.graphs.nodes import HexNodes
from anemoi.graphs.nodes import TriNodes


class TestMultiScaleEdgesInit:
    def test_init(self):
        """Test MultiScaleEdges initialization."""
        assert isinstance(MultiScaleEdges("test_nodes", "test_nodes", 1), MultiScaleEdges)

    @pytest.mark.parametrize("x_hops", [-0.5, "hello", None, -4])
    def test_fail_init(self, x_hops: str):
        """Test MultiScaleEdges initialization with invalid x_hops."""
        with pytest.raises(AssertionError):
            MultiScaleEdges("test_nodes", "test_nodes", x_hops)

    def test_fail_init_diff_nodes(self):
        """Test MultiScaleEdges initialization with invalid nodes."""
        with pytest.raises(AssertionError):
            MultiScaleEdges("test_nodes", "test_nodes2", 0)


class TestMultiScaleEdgesTransform:

    @pytest.fixture()
    def tri_ico_graph(self) -> HeteroData:
        """Return a HeteroData object with MultiScaleEdges."""
        graph = HeteroData()
        graph = TriNodes(1, "test_tri_nodes").update_graph(graph, {})
        graph["fail_nodes"].x = [1, 2, 3]
        graph["fail_nodes"].node_type = "FailNodes"
        return graph

    @pytest.fixture()
    def hex_ico_graph(self) -> HeteroData:
        """Return a HeteroData object with TriNodes."""
        graph = HeteroData()
        graph = HexNodes(1, "test_hex_nodes").update_graph(graph, {})
        graph["fail_nodes"].x = [1, 2, 3]
        graph["fail_nodes"].node_type = "FailNodes"
        return graph

    def test_transform_same_src_dst_tri_nodes(self, tri_ico_graph: HeteroData):
        """Test MultiScaleEdges update method."""

        edges = MultiScaleEdges("test_tri_nodes", "test_tri_nodes", 1)
        graph = edges.update_graph(tri_ico_graph)
        assert ("test_tri_nodes", "to", "test_tri_nodes") in graph.edge_types

    def test_transform_same_src_dst_hex_nodes(self, hex_ico_graph: HeteroData):
        """Test MultiScaleEdges update method."""

        edges = MultiScaleEdges("test_hex_nodes", "test_hex_nodes", 1)
        graph = edges.update_graph(hex_ico_graph)
        assert ("test_hex_nodes", "to", "test_hex_nodes") in graph.edge_types

    def test_transform_fail_nodes(self, tri_ico_graph: HeteroData):
        """Test MultiScaleEdges update method with wrong node type."""
        edges = MultiScaleEdges("fail_nodes", "fail_nodes", 1)
        with pytest.raises(AssertionError):
            edges.update_graph(tri_ico_graph)



import pytest
import torch
import zarr
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes import builder
from anemoi.graphs.nodes.attributes import AreaWeights
from anemoi.graphs.nodes.attributes import UniformWeights


def test_init(mocker, mock_zarr_dataset):
    """Test ZarrDatasetNodes initialization."""
    mocker.patch.object(builder, "open_dataset", return_value=mock_zarr_dataset)
    node_builder = builder.ZarrDatasetNodes("dataset.zarr", name="test_nodes")

    assert isinstance(node_builder, builder.BaseNodeBuilder)
    assert isinstance(node_builder, builder.ZarrDatasetNodes)


def test_fail_init():
    """Test ZarrDatasetNodes initialization with invalid resolution."""
    with pytest.raises(zarr.errors.PathNotFoundError):
        builder.ZarrDatasetNodes("invalid_path.zarr", name="test_nodes")


def test_register_nodes(mocker, mock_zarr_dataset):
    """Test ZarrDatasetNodes register correctly the nodes."""
    mocker.patch.object(builder, "open_dataset", return_value=mock_zarr_dataset)
    node_builder = builder.ZarrDatasetNodes("dataset.zarr", name="test_nodes")
    graph = HeteroData()

    graph = node_builder.register_nodes(graph)

    assert graph["test_nodes"].x is not None
    assert isinstance(graph["test_nodes"].x, torch.Tensor)
    assert graph["test_nodes"].x.shape == (node_builder.dataset.num_nodes, 2)
    assert graph["test_nodes"].node_type == "ZarrDatasetNodes"


@pytest.mark.parametrize("attr_class", [UniformWeights, AreaWeights])
def test_register_attributes(mocker, graph_with_nodes: HeteroData, attr_class):
    """Test ZarrDatasetNodes register correctly the weights."""
    mocker.patch.object(builder, "open_dataset", return_value=None)
    node_builder = builder.ZarrDatasetNodes("dataset.zarr", name="test_nodes")
    config = {"test_attr": {"_target_": f"anemoi.graphs.nodes.attributes.{attr_class.__name__}"}}

    graph = node_builder.register_attributes(graph_with_nodes, config)

    assert graph["test_nodes"]["test_attr"] is not None
    assert isinstance(graph["test_nodes"]["test_attr"], torch.Tensor)
    assert graph["test_nodes"]["test_attr"].shape[0] == graph["test_nodes"].x.shape[0]



import pytest

from anemoi.graphs.edges import CutOffEdges


def test_init():
    """Test CutOffEdges initialization."""
    CutOffEdges("test_nodes1", "test_nodes2", 0.5)


@pytest.mark.parametrize("cutoff_factor", [-0.5, "hello", None])
def test_fail_init(cutoff_factor: str):
    """Test CutOffEdges initialization with invalid cutoff."""
    with pytest.raises(AssertionError):
        CutOffEdges("test_nodes1", "test_nodes2", cutoff_factor)


def test_cutoff(graph_with_nodes):
    """Test CutOffEdges."""
    builder = CutOffEdges("test_nodes", "test_nodes", 0.5)
    graph = builder.update_graph(graph_with_nodes)
    assert ("test_nodes", "to", "test_nodes") in graph.edge_types



import logging
from abc import ABC
from abc import abstractmethod
from typing import Optional

import numpy as np
import torch
from scipy.spatial import SphericalVoronoi
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.generate.transforms import to_sphere_xyz
from anemoi.graphs.normalizer import NormalizerMixin

logger = logging.getLogger(__name__)


class BaseWeights(ABC, NormalizerMixin):
    """Base class for the weights of the nodes."""

    def __init__(self, norm: Optional[str] = None):
        self.norm = norm

    @abstractmethod
    def compute(self, nodes: NodeStorage, *args, **kwargs): ...

    def get_weights(self, *args, **kwargs) -> torch.Tensor:
        weights = self.compute(*args, **kwargs)
        if weights.ndim == 1:
            weights = weights[:, np.newaxis]
        norm_weights = self.normalize(weights)
        return torch.tensor(norm_weights, dtype=torch.float32)


class UniformWeights(BaseWeights):
    """Implements a uniform weight for the nodes."""

    def compute(self, nodes: NodeStorage) -> np.ndarray:
        return np.ones(nodes.num_nodes)


class AreaWeights(BaseWeights):
    """Implements the area of the nodes as the weights."""

    def __init__(self, norm: str = "unit-max", radius: float = 1.0, centre: np.ndarray = np.array([0, 0, 0])):
        super().__init__(norm=norm)

        # Weighting of the nodes
        self.radius = radius
        self.centre = centre

    def compute(self, nodes: NodeStorage, *args, **kwargs) -> np.ndarray:
        latitudes, longitudes = nodes.x[:, 0], nodes.x[:, 1]
        points = to_sphere_xyz((latitudes, longitudes))
        sv = SphericalVoronoi(points, self.radius, self.centre)
        area_weights = sv.calculate_areas()
        logger.debug(
            "There are %d of weights, which (unscaled) add up a total weight of %.2f.",
            len(area_weights),
            np.array(area_weights).sum(),
        )
        return area_weights



from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod
from pathlib import Path

import numpy as np
import torch
from anemoi.datasets import open_dataset
from anemoi.utils.config import DotDict
from hydra.utils import instantiate
from torch_geometric.data import HeteroData

from anemoi.graphs.generate.hexagonal import create_hexagonal_nodes
from anemoi.graphs.generate.icosahedral import create_icosahedral_nodes

LOGGER = logging.getLogger(__name__)


class BaseNodeBuilder(ABC):
    """Base class for node builders.

    The node coordinates are stored in the `x` attribute of the nodes and they are stored in radians.

    Attributes
    ----------
    name : str
        name of the nodes, key for the nodes in the HeteroData graph object.
    """

    def __init__(self, name: str) -> None:
        self.name = name

    def register_nodes(self, graph: HeteroData) -> None:
        """Register nodes in the graph.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the nodes.
        """
        graph[self.name].x = self.get_coordinates()
        graph[self.name].node_type = type(self).__name__
        return graph

    def register_attributes(self, graph: HeteroData, config: DotDict = None) -> HeteroData:
        """Register attributes in the nodes of the graph specified.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the attributes.
        config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with the registered attributes.
        """
        for attr_name, attr_config in config.items():
            graph[self.name][attr_name] = instantiate(attr_config).compute(graph, self.name)
        return graph

    @abstractmethod
    def get_coordinates(self) -> torch.Tensor: ...

    def reshape_coords(self, latitudes: np.ndarray, longitudes: np.ndarray) -> torch.Tensor:
        """Reshape latitude and longitude coordinates.

        Parameters
        ----------
        latitudes : np.ndarray of shape (N, )
            Latitude coordinates, in degrees.
        longitudes : np.ndarray of shape (N, )
            Longitude coordinates, in degrees.

        Returns
        -------
        torch.Tensor of shape (N, 2)
            A 2D tensor with the coordinates, in radians.
        """
        coords = np.stack([latitudes, longitudes], axis=-1).reshape((-1, 2))
        coords = np.deg2rad(coords)
        return torch.tensor(coords, dtype=torch.float32)

    def update_graph(self, graph: HeteroData, attr_config: DotDict | None = None) -> HeteroData:
        """Update the graph with new nodes.

        Parameters
        ----------
        graph : HeteroData
            Input graph.
        attr_config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with new nodes included.
        """
        graph = self.register_nodes(graph)

        if attr_config is None:
            return graph

        graph = self.register_attributes(graph, attr_config)

        return graph


class ZarrDatasetNodes(BaseNodeBuilder):
    """Nodes from Zarr dataset.

    Attributes
    ----------
    dataset : zarr.core.Array
        The dataset.

    Methods
    -------
    register_nodes(graph)
        Register the nodes in the graph.
    register_attributes(graph, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, attr_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(self, dataset: DotDict, name: str) -> None:
        LOGGER.info("Reading the dataset from %s.", dataset)
        self.dataset = open_dataset(dataset)
        super().__init__(name)

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (N, 2)
            Coordinates of the nodes.
        """
        return self.reshape_coords(self.dataset.latitudes, self.dataset.longitudes)


class NPZFileNodes(BaseNodeBuilder):
    """Nodes from NPZ defined grids.

    Attributes
    ----------
    resolution : str
        The resolution of the grid.
    grid_definition_path : str
        Path to the folder containing the grid definition files.
    grid_definition : dict[str, np.ndarray]
        The grid definition.

    Methods
    -------
    register_nodes(graph)
        Register the nodes in the graph.
    register_attributes(graph, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, attr_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(self, resolution: str, grid_definition_path: str, name: str) -> None:
        """Initialize the NPZFileNodes builder.

        The builder suppose the grids are stored in files with the name `grid-{resolution}.npz`.

        Parameters
        ----------
        resolution : str
            The resolution of the grid.
        grid_definition_path : str
            Path to the folder containing the grid definition files.
        """
        self.resolution = resolution
        self.grid_definition_path = grid_definition_path
        self.grid_definition = np.load(Path(self.grid_definition_path) / f"grid-{self.resolution}.npz")
        super().__init__(name)

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (N, 2)
            Coordinates of the nodes.
        """
        coords = self.reshape_coords(self.grid_definition["latitudes"], self.grid_definition["longitudes"])
        return coords


class IcosahedralNodes(BaseNodeBuilder, ABC):
    """Processor mesh based on a triangular mesh.

    It is based on the icosahedral mesh, which is a mesh of triangles that covers the sphere.

    Parameters
    ----------
    resolution : list[int] | int
        Refinement level of the mesh.
    """

    def __init__(
        self,
        resolution: int | list[int],
        name: str,
    ) -> None:
        self.resolutions = list(range(resolution + 1)) if isinstance(resolution, int) else resolution
        super().__init__(name)

    def get_coordinates(self) -> torch.Tensor:
        self.nx_graph, coords_rad, self.node_ordering = self.create_nodes()
        return torch.tensor(coords_rad[self.node_ordering], dtype=torch.float32)

    @abstractmethod
    def create_nodes(self) -> np.ndarray: ...

    def register_attributes(self, graph: HeteroData, config: DotDict) -> HeteroData:
        graph[self.name]["_resolutions"] = self.resolutions
        graph[self.name]["_nx_graph"] = self.nx_graph
        graph[self.name]["_node_ordering"] = self.node_ordering
        return super().register_attributes(graph, config)


class TriNodes(IcosahedralNodes):
    """Nodes based on iterative refinements of an icosahedron.

    It depends on the trimesh Python library.

    Attributes
    ----------
    resolutions : list[int]
        Refinement level of the mesh.
    name : str
        The name of the nodes.

    Methods
    -------
    register_nodes(graph)
        Register the nodes in the graph.
    register_attributes(graph, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, attr_config)
        Update the graph with new nodes and attributes.
    """

    def create_nodes(self) -> np.ndarray:
        return create_icosahedral_nodes(resolutions=self.resolutions)


class HexNodes(IcosahedralNodes):
    """Nodes based on iterative refinements of an icosahedron.

    It depends on the h3 Python library.

    Attributes
    ----------
    resolutions : list[int]
        Refinement level of the mesh.
    name : str
        The name of the nodes.

    Methods
    -------
    register_nodes(graph)
        Register the nodes in the graph.
    register_attributes(graph, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, attr_config)
        Update the graph with new nodes and attributes.
    """

    def create_nodes(self) -> np.ndarray:
        return create_hexagonal_nodes(self.resolutions)


class HEALPixNodes(BaseNodeBuilder):
    """Nodes from HEALPix grid.

    HEALPix is an acronym for Hierarchical Equal Area isoLatitude Pixelization of a sphere.

    Attributes
    ----------
    resolution : int
        The resolution of the grid.
    name : str
        The name of the nodes.

    Methods
    -------
    register_nodes(graph, name)
        Register the nodes in the graph.
    register_attributes(graph, name, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, name, attr_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(self, resolution: int, name: str) -> None:
        """Initialize the HEALPixNodes builder."""
        self.resolution = resolution
        super().__init__(name)

        assert isinstance(resolution, int), "Resolution must be an integer."
        assert resolution > 0, "Resolution must be positive."

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (N, 2)
            Coordinates of the nodes.
        """
        import healpy as hp

        spatial_res_degrees = hp.nside2resol(2**self.resolution, arcmin=True) / 60
        LOGGER.info(f"Creating HEALPix nodes with resolution {spatial_res_degrees:.2} deg.")

        npix = hp.nside2npix(2**self.resolution)
        hpxlon, hpxlat = hp.pix2ang(2**self.resolution, range(npix), nest=True, lonlat=True)

        return self.reshape_coords(hpxlat, hpxlon)



from __future__ import annotations

import logging
from itertools import chain
from pathlib import Path

import torch
from anemoi.utils.config import DotDict
from hydra.utils import instantiate
from torch_geometric.data import HeteroData

LOGGER = logging.getLogger(__name__)


class GraphCreator:
    """Graph creator."""

    def __init__(
        self,
        config: str | Path | DotDict,
    ):
        if isinstance(config, Path) or isinstance(config, str):
            self.config = DotDict.from_file(config)
        else:
            self.config = config

    def generate_graph(self) -> HeteroData:
        """Generate the graph.

        It instantiates the node builders and edge builders defined in the configuration
        file and applies them to the graph.

        Returns
        -------
            HeteroData: The generated graph.
        """
        graph = HeteroData()

        for nodes_name, nodes_cfg in self.config.nodes.items():
            graph = instantiate(nodes_cfg.node_builder, name=nodes_name).update_graph(
                graph, nodes_cfg.get("attributes", {})
            )

        for edges_cfg in self.config.edges:
            graph = instantiate(edges_cfg.edge_builder, edges_cfg.source_name, edges_cfg.target_name).update_graph(
                graph, edges_cfg.get("attributes", {})
            )

        return graph

    def clean(self, graph: HeteroData) -> HeteroData:
        """Remove private attributes used during creation from the graph.

        Parameters
        ----------
        graph : HeteroData
            generated graph

        Returns
        -------
        HeteroData
            cleaned graph
        """
        LOGGER.info("Cleaning graph.")
        for type_name in chain(graph.node_types, graph.edge_types):
            attr_names_to_remove = [attr_name for attr_name in graph[type_name] if attr_name.startswith("_")]
            for attr_name in attr_names_to_remove:
                del graph[type_name][attr_name]
                LOGGER.info(f"{attr_name} deleted from graph.")

        return graph

    def save(self, graph: HeteroData, save_path: Path, overwrite: bool = False) -> None:
        """Save the generated graph to the output path.

        Parameters
        ----------
        graph : HeteroData
            generated graph
        save_path : Path
            location to save the graph
        overwrite : bool, optional
            whether to overwrite existing graph file, by default False
        """
        save_path = Path(save_path)

        if not save_path.exists() or overwrite:
            save_path.parent.mkdir(parents=True, exist_ok=True)
            torch.save(graph, save_path)
            LOGGER.info(f"Graph saved at {save_path}.")
        else:
            LOGGER.info("Graph already exists. Use overwrite=True to overwrite.")

    def create(self, save_path: Path | None = None, overwrite: bool = False) -> HeteroData:
        """Create the graph and save it to the output path.

        Parameters
        ----------
        save_path : Path, optional
            location to save the graph, by default None
        overwrite : bool, optional
            whether to overwrite existing graph file, by default False

        Returns
        -------
        HeteroData
            created graph object
        """

        graph = self.generate_graph()
        graph = self.clean(graph)

        if save_path is None:
            LOGGER.warning("No output path specified. The graph will not be saved.")
        else:
            self.save(graph, save_path, overwrite)

        return graph



import numpy as np


def cartesian_to_latlon_degrees(xyz: np.ndarray) -> np.ndarray:
    """3D to lat-lon (in degrees) conversion.

    Convert 3D coordinates of points to the (lat, lon) on the sphere containing
    them.

    Parameters
    ----------
    xyz : np.ndarray
        The 3D coordinates of points.

    Returns
    -------
    np.ndarray
        A 2D array of lat-lon coordinates of shape (N, 2).
    """
    lat = np.arcsin(xyz[..., 2] / (xyz**2).sum(axis=1)) * 180.0 / np.pi
    lon = np.arctan2(xyz[..., 1], xyz[..., 0]) * 180.0 / np.pi
    return np.array((lat, lon), dtype=np.float32).transpose()


def cartesian_to_latlon_rad(xyz: np.ndarray) -> np.ndarray:
    """3D to lat-lon (in radians) conversion.

    Convert 3D coordinates of points to its coordinates on the sphere containing
    them.

    Parameters
    ----------
    xyz : np.ndarray
        The 3D coordinates of points.

    Returns
    -------
    np.ndarray
        A 2D array of the coordinates of shape (N, 2) in radians.
    """
    lat = np.arcsin(xyz[..., 2] / (xyz**2).sum(axis=1))
    lon = np.arctan2(xyz[..., 1], xyz[..., 0])
    return np.array((lat, lon), dtype=np.float32).transpose()


def latlon_rad_to_cartesian(loc: tuple[np.ndarray, np.ndarray], radius: float = 1) -> np.ndarray:
    """Convert planar coordinates to 3D coordinates in a sphere.

    Parameters
    ----------
    loc : np.ndarray
        The 2D coordinates of the points, in radians.
    radius : float, optional
        The radius of the sphere containing los points. Defaults to the unit sphere.

    Returns
    -------
    np.array of shape (3, num_points)
        3D coordinates of the points in the sphere.
    """
    latr, lonr = loc[0], loc[1]
    x = radius * np.cos(latr) * np.cos(lonr)
    y = radius * np.cos(latr) * np.sin(lonr)
    z = radius * np.sin(latr)
    return np.array((x, y, z)).T


def direction_vec(points: np.ndarray, reference: np.ndarray, epsilon: float = 10e-11) -> np.ndarray:
    """Direction vector computation.

    Compute the direction vector of a set of points with respect to a reference
    vector.

    Parameters
    ----------
    points : np.array of shape (num_points, 3)
        The points to compute the direction vector.
    reference : np.array of shape (3, )
        The reference vector.
    epsilon : float, optional
        The value to add to the first vector to avoid division by zero. Defaults to 10e-11.

    Returns
    -------
    np.array of shape (3, num_points)
        The direction vector of the cross product of the two vectors.
    """
    v = np.cross(points, reference)
    vnorm1 = np.power(v, 2).sum(axis=-1)
    redo_idx = np.where(vnorm1 < epsilon)[0]
    if len(redo_idx) > 0:
        points[redo_idx] += epsilon
        v = np.cross(points, reference)
        vnorm1 = np.power(v, 2).sum(axis=-1)
    return v.T / np.sqrt(vnorm1)



from __future__ import annotations

import h3
import networkx as nx
import numpy as np
from sklearn.metrics.pairwise import haversine_distances


def create_hexagonal_nodes(
    resolutions: list[int],
    area: dict | None = None,
) -> tuple[nx.Graph, np.ndarray, list[int]]:
    """Creates a global mesh from a refined icosahedro.

    This method relies on the H3 python library, which covers the earth with hexagons (and 5 pentagons). At each
    refinement level, a hexagon cell (nodes) has 7 child cells (aperture 7).

    Parameters
    ----------
    resolutions : list[int]
        Levels of mesh resolution to consider.
    area : dict
        A region, in GeoJSON data format, to be contained by all cells. Defaults to None, which computes the global
        mesh.

    Returns
    -------
    graph : networkx.Graph
        The specified graph (nodes & edges).
    coords_rad : np.ndarray
        The node coordinates (not ordered) in radians.
    node_ordering : list[int]
        Order of the nodes in the graph to be sorted by latitude and longitude.
    """
    graph = nx.Graph()

    area_kwargs = {"area": area}

    for resolution in resolutions:
        graph = add_nodes_for_resolution(graph, resolution, **area_kwargs)

    coords = np.deg2rad(np.array([h3.h3_to_geo(node) for node in graph.nodes]))

    # Sort nodes by latitude and longitude
    node_ordering = np.lexsort(coords.T[::-1], axis=0)

    return graph, coords, list(node_ordering)


def add_nodes_for_resolution(
    graph: nx.Graph,
    resolution: int,
    **area_kwargs: dict | None,
) -> nx.Graph:
    """Add all nodes at a specified refinement level to a graph.

    Parameters
    ----------
    graph : networkx.Graph
        The graph to add the nodes.
    resolution : int
        The H3 refinement level. It can be an integer from 0 to 15.
    area_kwargs: dict
        Additional arguments to pass to the get_nodes_at_resolution function.
    """

    nodes = get_nodes_at_resolution(resolution, **area_kwargs)

    for idx in nodes:
        graph.add_node(idx, hcoords_rad=np.deg2rad(h3.h3_to_geo(idx)))

    return graph


def get_nodes_at_resolution(
    resolution: int,
    area: dict | None = None,
) -> set[str]:
    """Get nodes at a specified refinement level over the entire globe.

    If area is not None, it will return the nodes within the specified area

    Parameters
    ----------
    resolution : int
        The H3 refinement level. It can be an integer from 0 to 15.
    area : dict
        An area as GeoJSON dictionary specifying a polygon. Defaults to None.

    Returns
    -------
    nodes : set[str]
        The set of H3 indexes at the specified resolution level.
    """
    nodes = h3.uncompact(h3.get_res0_indexes(), resolution) if area is None else h3.polyfill(area, resolution)

    # TODO: AOI not used in the current implementation.

    return nodes


def add_edges_to_nx_graph(
    graph: nx.Graph,
    resolutions: list[int],
    x_hops: int = 1,
    depth_children: int = 1,
) -> nx.Graph:
    """Adds the edges to the graph.

    This method includes multi-scale connections to the existing graph. The different scales
    are defined by the resolutions (or refinement levels) specified.

    Parameters
    ----------
    graph : networkx.Graph
        The graph to add the edges.
    resolutions : list[int]
        Levels of mesh resolution to consider.
    x_hops: int
        The number of hops to consider for the neighbours.
    depth_children : int
        The number of resolution levels to consider for the connections of children. Defaults to 1, which includes
        connections up to the next resolution level.

    Returns
    -------
    graph : networkx.Graph
        The graph with the added edges.
    """

    graph = add_neighbour_edges(graph, resolutions, x_hops)
    graph = add_edges_to_children(
        graph,
        resolutions,
        depth_children,
    )
    return graph


def add_neighbour_edges(
    graph: nx.Graph,
    refinement_levels: tuple[int],
    x_hops: int = 1,
) -> nx.Graph:
    for resolution in refinement_levels:
        nodes = select_nodes_from_graph_at_resolution(graph, resolution)

        for idx in nodes:
            # neighbours
            for idx_neighbour in h3.k_ring(idx, k=x_hops) & set(nodes):
                graph = add_edge(
                    graph,
                    h3.h3_to_center_child(idx, refinement_levels[-1]),
                    h3.h3_to_center_child(idx_neighbour, refinement_levels[-1]),
                )

    return graph


def add_edges_to_children(
    graph: nx.Graph,
    refinement_levels: tuple[int],
    depth_children: int | None = None,
) -> nx.Graph:
    """Adds edges to the children of the nodes at the specified resolution levels.

    Parameters
    ----------
    graph : nx.Graph
        graph to which the edges will be added
    refinement_levels : tuple[int]
        set of refinement levels
    depth_children : Optional[int], optional
        The number of resolution levels to consider for the connections of children. Defaults to 1, which includes
        connections up to the next resolution level, by default None
    """
    if depth_children is None:
        depth_children = len(refinement_levels)

    for i_level, resolution_parent in enumerate(refinement_levels[0:-1]):
        parent_nodes = select_nodes_from_graph_at_resolution(graph, resolution_parent)

        for parent_idx in parent_nodes:
            # add own children
            for resolution_child in refinement_levels[i_level + 1 : i_level + depth_children + 1]:
                for child_idx in h3.h3_to_children(parent_idx, res=resolution_child):
                    graph = add_edge(
                        graph,
                        h3.h3_to_center_child(parent_idx, refinement_levels[-1]),
                        h3.h3_to_center_child(child_idx, refinement_levels[-1]),
                    )

    return graph


def select_nodes_from_graph_at_resolution(graph: nx.Graph, resolution: int):
    parent_nodes = [node for node in graph.nodes if h3.h3_get_resolution(node) == resolution]
    return parent_nodes


def add_edge(
    graph: nx.Graph,
    source_node_h3_idx: str,
    target_node_h3_idx: str,
) -> nx.Graph:
    """Add edge between two nodes to a graph.

    The edge will only be added in case both target and source nodes are included in the graph.

    Parameters
    ----------
    graph : networkx.Graph
        The graph to add the nodes.
    source_node_h3_idx : str
        The H3 index of the tail of the edge.
    target_node_h3_idx : str
        The H3 index of the head of the edge.
    """
    if not graph.has_node(source_node_h3_idx) or not graph.has_node(target_node_h3_idx):
        return graph

    if source_node_h3_idx != target_node_h3_idx:
        source_location = np.deg2rad(h3.h3_to_geo(source_node_h3_idx))
        target_location = np.deg2rad(h3.h3_to_geo(target_node_h3_idx))
        graph.add_edge(
            source_node_h3_idx, target_node_h3_idx, weight=haversine_distances([source_location, target_location])[0][1]
        )

    return graph



import pytest

from anemoi.graphs.edges import KNNEdges


def test_init():
    """Test KNNEdges initialization."""
    KNNEdges("test_nodes1", "test_nodes2", 3)


@pytest.mark.parametrize("num_nearest_neighbours", [-1, 2.6, "hello", None])
def test_fail_init(num_nearest_neighbours: str):
    """Test KNNEdges initialization with invalid number of nearest neighbours."""
    with pytest.raises(AssertionError):
        KNNEdges("test_nodes1", "test_nodes2", num_nearest_neighbours)


def test_knn(graph_with_nodes):
    """Test KNNEdges."""
    builder = KNNEdges("test_nodes", "test_nodes", 3)
    graph = builder.update_graph(graph_with_nodes)
    assert ("test_nodes", "to", "test_nodes") in graph.edge_types



from __future__ import annotations

import numpy as np
import torch
from sklearn.neighbors import NearestNeighbors


def get_nearest_neighbour(coords_rad: torch.Tensor, mask: torch.Tensor | None = None) -> NearestNeighbors:
    """Get NearestNeighbour object fitted to coordinates.

    Parameters
    ----------
    coords_rad : torch.Tensor
        corrdinates in radians
    mask : Optional[torch.Tensor], optional
        mask to remove nodes, by default None

    Returns
    -------
    NearestNeighbors
        fitted NearestNeighbour object
    """
    assert mask is None or mask.shape == (
        coords_rad.shape[0],
        1,
    ), "Mask must have the same shape as the number of nodes."

    nearest_neighbour = NearestNeighbors(metric="haversine", n_jobs=4)

    nearest_neighbour.fit(coords_rad)

    return nearest_neighbour


def get_grid_reference_distance(coords_rad: torch.Tensor, mask: torch.Tensor | None = None) -> float:
    """Get the reference distance of the grid.

    It is the maximum distance of a node in the mesh with respect to its nearest neighbour.

    Parameters
    ----------
    coords_rad : torch.Tensor
        corrdinates in radians
    mask : Optional[torch.Tensor], optional
        mask to remove nodes, by default None

    Returns
    -------
    float
        The reference distance of the grid.
    """
    nearest_neighbours = get_nearest_neighbour(coords_rad, mask)
    dists, _ = nearest_neighbours.kneighbors(coords_rad, n_neighbors=2, return_distance=True)
    return dists[dists > 0].max()


def add_margin(lats: np.ndarray, lons: np.ndarray, margin: float) -> tuple[np.ndarray, np.ndarray]:
    """Add a margin to the convex hull of the points considered.

    For each point (lat, lon) add 8 points around it, each at a distance of `margin` from the original point.

    Arguments
    ---------
    lats : np.ndarray
        Latitudes of the points considered.
    lons : np.ndarray
        Longitudes of the points considered.
    margin : float
        The margin to add to the convex hull.

    Returns
    -------
    latitudes : np.ndarray
        Latitudes of the points considered, including the margin.
    longitudes : np.ndarray
        Longitudes of the points considered, including the margin.
    """
    assert margin >= 0, "Margin must be non-negative"
    if margin == 0:
        return lats, lons

    latitudes, longitudes = [], []
    for lat_sign in [-1, 0, 1]:
        for lon_sign in [-1, 0, 1]:
            latitudes.append(lats + lat_sign * margin)
            longitudes.append(lons + lon_sign * margin)

    return np.concatenate(latitudes), np.concatenate(longitudes)


def get_index_in_outer_join(vector: torch.Tensor, tensor: torch.Tensor) -> int:
    """Index position of vector.

    Get the index position of a vector in a matrix.

    Parameters
    ----------
    vector : torch.Tensor of shape (N, )
        Vector to get its position in the matrix.
    tensor : torch.Tensor of shape (M, N,)
        Tensor in which the position is searched.

    Returns
    -------
    int
        Index position of `vector` in `tensor`. -1 if `vector` is not in `tensor`.
    """
    mask = torch.all(tensor == vector, axis=1)
    if mask.any():
        return int(torch.where(mask)[0])
    return -1


def haversine_distance(source_coords: np.ndarray, target_coords: np.ndarray) -> np.ndarray:
    """Haversine distance.

    Parameters
    ----------
    source_coords : np.ndarray of shape (N, 2)
        Source coordinates in radians.
    target_coords : np.ndarray of shape (N, 2)
        Destination coordinates in radians.

    Returns
    -------
    np.ndarray of shape (N,)
        Haversine distance between source and destination coordinates.
    """
    dlat = target_coords[:, 0] - source_coords[:, 0]
    dlon = target_coords[:, 1] - source_coords[:, 1]
    a = np.sin(dlat / 2) ** 2 + np.cos(source_coords[:, 0]) * np.cos(target_coords[:, 0]) * np.sin(dlon / 2) ** 2
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
    return c



from anemoi.graphs.descriptor import GraphDescriptor

from . import Command


class Describe(Command):
    """Describe a graph."""

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument("graph_file", help="Path to the graph (a .PT file).")

    def run(self, args):
        kwargs = vars(args)

        GraphDescriptor(kwargs["graph_file"]).describe()


command = Describe



import pytest
import torch

from anemoi.graphs.edges.attributes import EdgeDirection
from anemoi.graphs.edges.attributes import EdgeLength


@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-std"])
@pytest.mark.parametrize("luse_rotated_features", [True, False])
def test_directional_features(graph_nodes_and_edges, norm, luse_rotated_features: bool):
    """Test EdgeDirection compute method."""
    edge_attr_builder = EdgeDirection(norm=norm, luse_rotated_features=luse_rotated_features)
    edge_attr = edge_attr_builder.compute(graph_nodes_and_edges, ("test_nodes", "to", "test_nodes"))
    assert isinstance(edge_attr, torch.Tensor)


@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-std"])
def test_edge_lengths(graph_nodes_and_edges, norm):
    """Test EdgeLength compute method."""
    edge_attr_builder = EdgeLength(norm=norm)
    edge_attr = edge_attr_builder.compute(graph_nodes_and_edges, ("test_nodes", "to", "test_nodes"))
    assert isinstance(edge_attr, torch.Tensor)


@pytest.mark.parametrize("attribute_builder", [EdgeDirection(), EdgeLength()])
def test_fail_edge_features(attribute_builder, graph_nodes_and_edges):
    """Test edge attribute builder fails with unknown nodes."""
    with pytest.raises(AssertionError):
        attribute_builder.compute(graph_nodes_and_edges, ("test_nodes", "to", "unknown_nodes"))



from typing import Optional

import numpy as np
import torch
from torch_geometric.data import HeteroData


def node_list(graph: HeteroData, nodes_name: str, mask: Optional[list[bool]] = None) -> tuple[list[float], list[float]]:
    """Get the latitude and longitude of the nodes.

    Parameters
    ----------
    graph : dict[str, torch.Tensor]
        Graph to plot.
    nodes_name : str
        Name of the nodes.
    mask : list[bool], optional
        Mask to filter the nodes. Default is None.

    Returns
    -------
    latitudes : list[float]
        Latitude coordinates of the nodes.
    longitudes : list[float]
        Longitude coordinates of the nodes.
    """
    coords = np.rad2deg(graph[nodes_name].x.numpy())
    latitudes = coords[:, 0]
    longitudes = coords[:, 1]
    if mask is not None:
        latitudes = latitudes[mask]
        longitudes = longitudes[mask]
    return latitudes.tolist(), longitudes.tolist()


def edge_list(graph: HeteroData, source_nodes_name: str, target_nodes_name: str) -> tuple[np.ndarray, np.ndarray]:
    """Get the edge list.

    This method returns the edge list to be represented in a graph. It computes the coordinates of the points connected
    and include NaNs to separate the edges.

    Parameters
    ----------
    graph : HeteroData
        Graph to plot.
    source_nodes_name : str
        Name of the source nodes.
    target_nodes_name : str
        Name of the target nodes.

    Returns
    -------
    latitudes : np.ndarray
        Latitude coordinates of the points connected.
    longitudes : np.ndarray
        Longitude coordinates of the points connected.
    """
    sub_graph = graph[(source_nodes_name, "to", target_nodes_name)].edge_index
    x0 = np.rad2deg(graph[source_nodes_name].x[sub_graph[0]])
    y0 = np.rad2deg(graph[target_nodes_name].x[sub_graph[1]])
    nans = np.full_like(x0[:, :1], np.nan)
    latitudes = np.concatenate([x0[:, :1], y0[:, :1], nans], axis=1).flatten()
    longitudes = np.concatenate([x0[:, 1:2], y0[:, 1:2], nans], axis=1).flatten()
    return latitudes, longitudes


def compute_node_adjacencies(
    graph: HeteroData, source_nodes_name: str, target_nodes_name: str
) -> tuple[list[int], list[str]]:
    """Compute the number of adjacencies of each target node in a bipartite graph.

    Parameters
    ----------
    graph : HeteroData
        Graph to plot.
    source_nodes_name : str
        Name of the dimension of the coordinates for the head nodes.
    target_nodes_name : str
        Name of the dimension of the coordinates for the tail nodes.

    Returns
    -------
    num_adjacencies : np.ndarray
        Number of adjacencies of each node.
    """
    node_adjacencies = np.zeros(graph[target_nodes_name].num_nodes, dtype=int)
    vals, counts = np.unique(graph[(source_nodes_name, "to", target_nodes_name)].edge_index[1], return_counts=True)
    node_adjacencies[vals] = counts
    return node_adjacencies


def get_node_adjancency_attributes(graph: HeteroData) -> dict[str, tuple[str, np.ndarray]]:
    """Get the node adjacencies for each subgraph."""
    node_adj_attr = {}
    for (source_nodes_name, _, target_nodes_name), _ in graph.edge_items():
        attr_name = f"# connections from {source_nodes_name}"
        node_adj_vector = compute_node_adjacencies(graph, source_nodes_name, target_nodes_name)
        if target_nodes_name not in node_adj_attr:
            node_adj_attr[target_nodes_name] = {attr_name: node_adj_vector}
        else:
            node_adj_attr[target_nodes_name][attr_name] = node_adj_vector

    return node_adj_attr


def compute_isolated_nodes(graph: HeteroData) -> dict[str, tuple[list, list]]:
    """Compute the isolated nodes.

    Parameters
    ----------
    graph : HeteroData
        Graph.

    Returns
    -------
    dict[str, list[int]]
        Dictionary with the isolated nodes for each subgraph.
    """
    isolated_nodes = {}
    for (source_name, _, target_name), sub_graph in graph.edge_items():
        head_isolated = np.ones(graph[source_name].num_nodes, dtype=bool)
        tail_isolated = np.ones(graph[target_name].num_nodes, dtype=bool)
        head_isolated[sub_graph.edge_index[0]] = False
        tail_isolated[sub_graph.edge_index[1]] = False
        if np.any(head_isolated):
            isolated_nodes[f"{source_name} isolated (--> {target_name})"] = node_list(
                graph, source_name, mask=list(head_isolated)
            )
        if np.any(tail_isolated):
            isolated_nodes[f"{target_name} isolated ({source_name} -->)"] = node_list(
                graph, target_name, mask=list(tail_isolated)
            )

    return isolated_nodes


def get_node_attribute_dims(graph: HeteroData) -> dict[str, int]:
    """Get dimensions of the node attributes.

    Parameters
    ----------
    graph : HeteroData
        The graph to inspect.

    Returns
    -------
    dict[str, int]
        A dictionary with the attribute names as keys and the number of dimensions as values.
    """
    attr_dims = {}
    for nodes in graph.node_stores:
        for attr in nodes.node_attrs():
            if attr == "x" or not isinstance(nodes[attr], torch.Tensor):
                continue
            elif attr not in attr_dims:
                attr_dims[attr] = nodes[attr].shape[1]
            else:
                assert (
                    nodes[attr].shape[1] == attr_dims[attr]
                ), f"Attribute {attr} has different dimensions in different nodes."
    return attr_dims


def get_edge_attribute_dims(graph: HeteroData) -> dict[str, int]:
    """Get dimensions of the node attributes.

    Parameters
    ----------
    graph : HeteroData
        The graph to inspect.

    Returns
    -------
    dict[str, int]
        A dictionary with the attribute names as keys and the number of dimensions as values.
    """
    attr_dims = {}
    for edges in graph.edge_stores:
        for attr in edges.edge_attrs():
            if attr == "edge_index":
                continue
            elif attr not in attr_dims:
                attr_dims[attr] = edges[attr].shape[1]
            else:
                assert (
                    edges[attr].shape[1] == attr_dims[attr]
                ), f"Attribute {attr} has different dimensions in different edges."
    return attr_dims



import logging

import numpy as np

LOGGER = logging.getLogger(__name__)


class NormalizerMixin:
    """Mixin class for normalizing attributes."""

    def normalize(self, values: np.ndarray) -> np.ndarray:
        """Normalize the given values.

        It supports different normalization methods: None, 'l1',
        'l2', 'unit-max' and 'unit-std'.

        Parameters
        ----------
        values : np.ndarray of shape (N, M)
            Values to normalize.

        Returns
        -------
        np.ndarray
            Normalized values.
        """
        if self.norm is None:
            LOGGER.debug(f"{self.__class__.__name__} values are not normalized.")
            return values
        if self.norm == "l1":
            return values / np.sum(values)
        if self.norm == "l2":
            return values / np.linalg.norm(values)
        if self.norm == "unit-max":
            return values / np.amax(values)
        if self.norm == "unit-std":
            std = np.std(values)
            if std == 0:
                LOGGER.warning(f"Std. dev. of the {self.__class__.__name__} values is 0. Normalization is skipped.")
                return values
            return values / std
        raise ValueError(
            f"Attribute normalization \"{self.norm}\" is not valid. Options are: 'l1', 'l2', 'unit-max' or 'unit-std'."
        )



import logging
import os
from pathlib import Path
from typing import Optional
from typing import Union

import torch

from anemoi.graphs.plotting.displots import plot_distribution_edge_attributes
from anemoi.graphs.plotting.displots import plot_distribution_node_attributes
from anemoi.graphs.plotting.displots import plot_distribution_node_derived_attributes
from anemoi.graphs.plotting.interactive_html import plot_interactive_nodes
from anemoi.graphs.plotting.interactive_html import plot_interactive_subgraph
from anemoi.graphs.plotting.interactive_html import plot_isolated_nodes

LOGGER = logging.getLogger(__name__)


class GraphInspector:
    """Inspect the graph."""

    def __init__(
        self,
        path: Union[str, Path],
        output_path: Path,
        show_attribute_distributions: Optional[bool] = True,
        show_nodes: Optional[bool] = False,
        **kwargs,
    ):
        self.path = path
        self.graph = torch.load(self.path)
        self.output_path = output_path
        self.show_attribute_distributions = show_attribute_distributions
        self.show_nodes = show_nodes

        if isinstance(self.output_path, str):
            self.output_path = Path(self.output_path)

        os.makedirs(self.output_path, exist_ok=True)

        assert self.output_path.is_dir(), f"Path {self.output_path} is not a directory."
        assert os.access(self.output_path, os.W_OK), f"Path {self.output_path} is not writable."

    def inspect(self):
        """Run all the inspector methods."""
        LOGGER.info("Saving interactive plots of isolated nodes ...")
        plot_isolated_nodes(self.graph, self.output_path / "isolated_nodes.html")

        LOGGER.info("Saving interactive plots of subgraphs ...")
        for edges_subgraph in self.graph.edge_types:
            ofile = self.output_path / f"{edges_subgraph[0]}_to_{edges_subgraph[2]}.html"
            plot_interactive_subgraph(self.graph, edges_subgraph, out_file=ofile)

        if self.show_attribute_distributions:
            LOGGER.info("Saving distribution plots of node ande edge attributes ...")
            plot_distribution_node_derived_attributes(self.graph, self.output_path / "distribution_node_adjancency.png")
            plot_distribution_edge_attributes(self.graph, self.output_path / "distribution_edge_attributes.png")
            plot_distribution_node_attributes(self.graph, self.output_path / "distribution_node_attributes.png")

        if self.show_nodes:
            LOGGER.info("Saving interactive plots of nodes ...")
            for nodes_name in self.graph.node_types:
                plot_interactive_nodes(self.graph, nodes_name, out_file=self.output_path / f"{nodes_name}_nodes.html")



from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod

import networkx as nx
import numpy as np
import torch
from anemoi.utils.config import DotDict
from hydra.utils import instantiate
from sklearn.neighbors import NearestNeighbors
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs import EARTH_RADIUS
from anemoi.graphs.generate import hexagonal
from anemoi.graphs.generate import icosahedral
from anemoi.graphs.nodes.builder import HexNodes
from anemoi.graphs.nodes.builder import TriNodes
from anemoi.graphs.utils import get_grid_reference_distance

LOGGER = logging.getLogger(__name__)


class BaseEdgeBuilder(ABC):
    """Base class for edge builders."""

    def __init__(self, source_name: str, target_name: str):
        self.source_name = source_name
        self.target_name = target_name

    @property
    def name(self) -> tuple[str, str, str]:
        """Name of the edge subgraph."""
        return self.source_name, "to", self.target_name

    @abstractmethod
    def get_adjacency_matrix(self, source_nodes: NodeStorage, target_nodes: NodeStorage): ...

    def prepare_node_data(self, graph: HeteroData) -> tuple[NodeStorage, NodeStorage]:
        """Prepare node information and get source and target nodes."""
        return graph[self.source_name], graph[self.target_name]

    def get_edge_index(self, graph: HeteroData) -> torch.Tensor:
        """Get the edge indices of source and target nodes.

        Parameters
        ----------
        graph : HeteroData
            The graph.

        Returns
        -------
        torch.Tensor of shape (2, num_edges)
            The edge indices.
        """
        source_nodes, target_nodes = self.prepare_node_data(graph)

        adjmat = self.get_adjacency_matrix(source_nodes, target_nodes)

        # Get source & target indices of the edges
        edge_index = np.stack([adjmat.col, adjmat.row], axis=0)

        return torch.from_numpy(edge_index).to(torch.int32)

    def register_edges(self, graph: HeteroData) -> HeteroData:
        """Register edges in the graph.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the edges.

        Returns
        -------
        HeteroData
            The graph with the registered edges.
        """
        graph[self.name].edge_index = self.get_edge_index(graph)
        graph[self.name].edge_type = type(self).__name__
        return graph

    def register_attributes(self, graph: HeteroData, config: DotDict) -> HeteroData:
        """Register attributes in the edges of the graph specified.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the attributes.
        config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with the registered attributes.
        """
        for attr_name, attr_config in config.items():
            graph[self.name][attr_name] = instantiate(attr_config).compute(graph, self.name)
        return graph

    def update_graph(self, graph: HeteroData, attrs_config: DotDict | None = None) -> HeteroData:
        """Update the graph with the edges.

        Parameters
        ----------
        graph : HeteroData
            The graph.
        attrs_config : DotDict
            The configuration of the edge attributes.

        Returns
        -------
        HeteroData
            The graph with the edges.
        """
        graph = self.register_edges(graph)

        if attrs_config is None:
            return graph

        graph = self.register_attributes(graph, attrs_config)

        return graph


class KNNEdges(BaseEdgeBuilder):
    """Computes KNN based edges and adds them to the graph.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    num_nearest_neighbours : int
        Number of nearest neighbours.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def __init__(self, source_name: str, target_name: str, num_nearest_neighbours: int):
        super().__init__(source_name, target_name)
        assert isinstance(num_nearest_neighbours, int), "Number of nearest neighbours must be an integer"
        assert num_nearest_neighbours > 0, "Number of nearest neighbours must be positive"
        self.num_nearest_neighbours = num_nearest_neighbours

    def get_adjacency_matrix(self, source_nodes: np.ndarray, target_nodes: np.ndarray):
        """Compute the adjacency matrix for the KNN method.

        Parameters
        ----------
        source_nodes : np.ndarray
            The source nodes.
        target_nodes : np.ndarray
            The target nodes.
        """
        assert self.num_nearest_neighbours is not None, "number of neighbors required for knn encoder"
        LOGGER.info(
            "Using KNN-Edges (with %d nearest neighbours) between %s and %s.",
            self.num_nearest_neighbours,
            self.source_name,
            self.target_name,
        )

        nearest_neighbour = NearestNeighbors(metric="haversine", n_jobs=4)
        nearest_neighbour.fit(source_nodes.x.numpy())
        adj_matrix = nearest_neighbour.kneighbors_graph(
            target_nodes.x.numpy(),
            n_neighbors=self.num_nearest_neighbours,
            mode="distance",
        ).tocoo()
        return adj_matrix


class CutOffEdges(BaseEdgeBuilder):
    """Computes cut-off based edges and adds them to the graph.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    cutoff_factor : float
        Factor to multiply the grid reference distance to get the cut-off radius.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def __init__(self, source_name: str, target_name: str, cutoff_factor: float):
        super().__init__(source_name, target_name)
        assert isinstance(cutoff_factor, (int, float)), "Cutoff factor must be a float"
        assert cutoff_factor > 0, "Cutoff factor must be positive"
        self.cutoff_factor = cutoff_factor

    def get_cutoff_radius(self, graph: HeteroData, mask_attr: torch.Tensor | None = None):
        """Compute the cut-off radius.

        The cut-off radius is computed as the product of the target nodes
        reference distance and the cut-off factor.

        Parameters
        ----------
        graph : HeteroData
            The graph.
        mask_attr : torch.Tensor
            The mask attribute.

        Returns
        -------
        float
            The cut-off radius.
        """
        target_nodes = graph[self.target_name]
        mask = target_nodes[mask_attr] if mask_attr is not None else None
        target_grid_reference_distance = get_grid_reference_distance(target_nodes.x, mask)
        radius = target_grid_reference_distance * self.cutoff_factor
        return radius

    def prepare_node_data(self, graph: HeteroData) -> tuple[NodeStorage, NodeStorage]:
        """Prepare node information and get source and target nodes."""
        self.radius = self.get_cutoff_radius(graph)
        return super().prepare_node_data(graph)

    def get_adjacency_matrix(self, source_nodes: NodeStorage, target_nodes: NodeStorage):
        """Get the adjacency matrix for the cut-off method.

        Parameters
        ----------
        source_nodes : NodeStorage
            The source nodes.
        target_nodes : NodeStorage
            The target nodes.
        """
        LOGGER.info(
            "Using CutOff-Edges (with radius = %.1f km) between %s and %s.",
            self.radius * EARTH_RADIUS,
            self.source_name,
            self.target_name,
        )

        nearest_neighbour = NearestNeighbors(metric="haversine", n_jobs=4)
        nearest_neighbour.fit(source_nodes.x)
        adj_matrix = nearest_neighbour.radius_neighbors_graph(target_nodes.x, radius=self.radius).tocoo()
        return adj_matrix


class MultiScaleEdges(BaseEdgeBuilder):
    """Base class for multi-scale edges in the nodes of a graph.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    x_hops : int
        Number of hops (in the refined icosahedron) between two nodes to connect
        them with an edge.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def __init__(self, source_name: str, target_name: str, x_hops: int):
        super().__init__(source_name, target_name)
        assert source_name == target_name, f"{self.__class__.__name__} requires source and target nodes to be the same."
        assert isinstance(x_hops, int), "Number of x_hops must be an integer"
        assert x_hops > 0, "Number of x_hops must be positive"
        self.x_hops = x_hops

    def adjacency_from_tri_nodes(self, source_nodes: NodeStorage):
        source_nodes["_nx_graph"] = icosahedral.add_edges_to_nx_graph(
            source_nodes["_nx_graph"],
            resolutions=source_nodes["_resolutions"],
            x_hops=self.x_hops,
        )  # HeteroData refuses to accept None

        adjmat = nx.to_scipy_sparse_array(
            source_nodes["_nx_graph"], nodelist=list(range(len(source_nodes["_nx_graph"]))), format="coo"
        )
        return adjmat

    def adjacency_from_hex_nodes(self, source_nodes: NodeStorage):

        source_nodes["_nx_graph"] = hexagonal.add_edges_to_nx_graph(
            source_nodes["_nx_graph"],
            resolutions=source_nodes["_resolutions"],
            x_hops=self.x_hops,
        )

        adjmat = nx.to_scipy_sparse_array(source_nodes["_nx_graph"], format="coo")
        return adjmat

    def get_adjacency_matrix(self, source_nodes: NodeStorage, target_nodes: NodeStorage):
        if self.node_type == TriNodes.__name__:
            adjmat = self.adjacency_from_tri_nodes(source_nodes)
        elif self.node_type == HexNodes.__name__:
            adjmat = self.adjacency_from_hex_nodes(source_nodes)
        else:
            raise ValueError(f"Invalid node type {self.node_type}")

        adjmat = self.post_process_adjmat(source_nodes, adjmat)

        return adjmat

    def post_process_adjmat(self, nodes: NodeStorage, adjmat):
        graph_sorted = {node_pos: i for i, node_pos in enumerate(nodes["_node_ordering"])}
        sort_func = np.vectorize(graph_sorted.get)
        adjmat.row = sort_func(adjmat.row)
        adjmat.col = sort_func(adjmat.col)
        return adjmat

    def update_graph(self, graph: HeteroData, attrs_config: DotDict | None = None) -> HeteroData:
        assert (
            graph[self.source_name].node_type == TriNodes.__name__
            or graph[self.source_name].node_type == HexNodes.__name__
        ), f"{self.__class__.__name__} requires {TriNodes.__name__} or {HexNodes.__name__}."

        self.node_type = graph[self.source_name].node_type

        return super().update_graph(graph, attrs_config)



from __future__ import annotations

from collections.abc import Iterable

import networkx as nx
import numpy as np
import trimesh
from sklearn.metrics.pairwise import haversine_distances
from sklearn.neighbors import BallTree

from anemoi.graphs.generate.transforms import cartesian_to_latlon_rad


def create_icosahedral_nodes(
    resolutions: list[int],
) -> tuple[nx.DiGraph, np.ndarray, list[int]]:
    """Creates a global mesh following AIFS strategy.

    This method relies on the trimesh python library.

    Parameters
    ----------
    resolutions : list[int]
        Levels of mesh resolution to consider.
    aoi_mask_builder : KNNAreaMaskBuilder
        KNNAreaMaskBuilder with the cloud of points to limit the mesh area, by default None.

    Returns
    -------
    graph : networkx.Graph
        The specified graph (nodes & edges).
    coords_rad : np.ndarray
        The node coordinates (not ordered) in radians.
    node_ordering : list[int]
        Order of the nodes in the graph to be sorted by latitude and longitude.
    """
    sphere = trimesh.creation.icosphere(subdivisions=resolutions[-1], radius=1.0)

    coords_rad = cartesian_to_latlon_rad(sphere.vertices)

    node_ordering = get_node_ordering(coords_rad)

    # TODO: AOI mask builder is not used in the current implementation.

    nx_graph = create_icosahedral_nx_graph_from_coords(coords_rad, node_ordering)

    return nx_graph, coords_rad, list(node_ordering)


def create_icosahedral_nx_graph_from_coords(coords_rad: np.ndarray, node_ordering: list[int]):

    graph = nx.DiGraph()
    for i, coords in enumerate(coords_rad[node_ordering]):
        node_id = node_ordering[i]
        graph.add_node(node_id, hcoords_rad=coords)

    assert list(graph.nodes.keys()) == list(node_ordering), "Nodes are not correctly added to the graph."
    assert graph.number_of_nodes() == len(node_ordering), "The number of nodes must be the same."
    return graph


def get_node_ordering(coords_rad: np.ndarray) -> np.ndarray:
    """Get the node ordering to sort the nodes by latitude and longitude."""
    # Get indices to sort points by lon & lat in radians.
    index_latitude = np.argsort(coords_rad[:, 1])
    index_longitude = np.argsort(coords_rad[index_latitude][:, 0])[::-1]
    node_ordering = np.arange(coords_rad.shape[0])[index_latitude][index_longitude]
    return node_ordering


def add_edges_to_nx_graph(
    graph: nx.DiGraph,
    resolutions: list[int],
    x_hops: int = 1,
) -> nx.DiGraph:
    """Adds the edges to the graph.

    This method adds multi-scale connections to the existing graph. The corresponfing nodes or vertices
    are defined by an isophere at the different esolutions (or refinement levels) specified.

    Parameters
    ----------
    graph : nx.DiGraph
        The graph to add the edges. It should correspond to the mesh nodes, without edges.
    resolutions : list[int]
        Levels of mesh refinement levels to consider.
    x_hops : int, optional
        Number of hops between 2 nodes to consider them neighbours, by default 1.
    aoi_mask_builder : KNNAreaMaskBuilder
        NearestNeighbors with the cloud of points to limit the mesh area, by default None.
    margin_radius_km : float, optional
        Margin radius in km to consider when creating the processor mesh, by default 0.0.

    Returns
    -------
    graph : nx.DiGraph
        The graph with the added edges.
    """
    assert x_hops > 0, "x_hops == 0, graph would have no edges ..."

    sphere = trimesh.creation.icosphere(subdivisions=resolutions[-1], radius=1.0)
    vertices_rad = cartesian_to_latlon_rad(sphere.vertices)
    node_neighbours = get_neighbours_within_hops(sphere, x_hops)

    for idx_node, idx_neighbours in node_neighbours.items():
        add_neigbours_edges(graph, vertices_rad, idx_node, idx_neighbours)

    tree = BallTree(vertices_rad, metric="haversine")

    # Build the multi-scale connections
    for resolution in resolutions[:-1]:
        # Define the isophere at specified 'resolution' level
        r_sphere = trimesh.creation.icosphere(subdivisions=resolution, radius=1.0)

        # Get the vertices of the isophere
        r_vertices_rad = cartesian_to_latlon_rad(r_sphere.vertices)

        # TODO AOI mask builder is not used in the current implementation.

        node_neighbours = get_neighbours_within_hops(r_sphere, x_hops)

        _, vertex_mapping_index = tree.query(r_vertices_rad, k=1)
        for idx_node, idx_neighbours in node_neighbours.items():
            add_neigbours_edges(
                graph, r_vertices_rad, idx_node, idx_neighbours, vertex_mapping_index=vertex_mapping_index
            )

    return graph


def get_neighbours_within_hops(tri_mesh: trimesh.Trimesh, x_hops: int) -> dict[int, set[int]]:
    """Get the neigbour connections in the graph.

    Parameters
    ----------
    tri_mesh : trimesh.Trimesh
        The mesh to consider.
    x_hops : int
        Number of hops between 2 nodes to consider them neighbours.
    valid_nodes : list[int], optional
        List of valid nodes to consider, by default None. It is useful to consider only a subset of the nodes to save
        computation time.

    Returns
    -------
    neighbours : dict[int, set[int]]
        A list with the neighbours for each vertex. The element at position 'i' correspond to the neighbours to the
        i-th vertex of the mesh.
    """
    edges = tri_mesh.edges_unique

    valid_nodes = list(range(len(tri_mesh.vertices)))
    graph = nx.from_edgelist(edges)

    # Get a dictionary of the neighbours within 'x_hops' neighbourhood of each node in the graph
    neighbours = {
        i: set(nx.ego_graph(graph, i, radius=x_hops, center=False) if i in graph else []) for i in valid_nodes
    }

    return neighbours


def add_neigbours_edges(
    graph: nx.Graph,
    vertices: np.ndarray,
    node_idx: int,
    neighbour_indices: Iterable[int],
    self_loops: bool = False,
    vertex_mapping_index: np.ndarray | None = None,
) -> None:
    """Adds the edges of one node to its neighbours.

    Parameters
    ----------
    graph : nx.Graph
        The graph.
    vertices : np.ndarray
        A 2D array of shape (num_vertices, 2) with the planar coordinates of the mesh, in radians.
    node_idx : int
        The node considered.
    neighbours : list[int]
        The neighbours of the node.
    self_loops : bool, optional
        Whether is supported to add self-loops, by default False.
    vertex_mapping_index : np.ndarray, optional
        Index to map the vertices from the refined sphere to the original one, by default None.
    """
    for neighbour_idx in neighbour_indices:
        if not self_loops and node_idx == neighbour_idx:  # no self-loops
            continue

        location_node = vertices[node_idx]
        location_neighbour = vertices[neighbour_idx]
        edge_length = haversine_distances([location_neighbour, location_node])[0][1]

        if vertex_mapping_index is not None:
            # Use the same method to add edge in all spheres
            node_neighbour = vertex_mapping_index[neighbour_idx][0]
            node = vertex_mapping_index[node_idx][0]
        else:
            node, node_neighbour = node_idx, neighbour_idx

        # add edge to the graph (if both source and target nodes are in the graph)
        if node in graph and node_neighbour in graph:
            graph.add_edge(node_neighbour, node, weight=edge_length)



from __future__ import annotations

from abc import ABC
from abc import abstractmethod

import numpy as np
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.edges.directional import directional_edge_features
from anemoi.graphs.normalizer import NormalizerMixin
from anemoi.graphs.utils import haversine_distance


class BaseEdgeAttribute(ABC, NormalizerMixin):
    """Base class for edge attributes."""

    def __init__(self, norm: str | None = None) -> None:
        self.norm = norm

    @abstractmethod
    def get_raw_values(self, graph: HeteroData, source_name: str, target_name: str, *args, **kwargs) -> np.ndarray: ...

    def post_process(self, values: np.ndarray) -> torch.Tensor:
        """Post-process the values."""
        if values.ndim == 1:
            values = values[:, np.newaxis]

        normed_values = self.normalize(values)

        return torch.tensor(normed_values, dtype=torch.float32)

    def compute(self, graph: HeteroData, edges_name: tuple[str, str, str], *args, **kwargs) -> torch.Tensor:
        """Compute the edge attributes."""
        source_name, _, target_name = edges_name
        assert (
            source_name in graph.node_types
        ), f"Node \"{source_name}\" not found in graph. Optional nodes are {', '.join(graph.node_types)}."
        assert (
            target_name in graph.node_types
        ), f"Node \"{target_name}\" not found in graph. Optional nodes are {', '.join(graph.node_types)}."

        values = self.get_raw_values(graph, source_name, target_name, *args, **kwargs)
        return self.post_process(values)


class EdgeDirection(BaseEdgeAttribute):
    """Edge direction feature.

    This class calculates the direction of an edge using either:
    1. Rotated features: The target nodes are rotated to the north pole to compute the edge direction.
    2. Non-rotated features: The direction is computed as the difference in latitude and longitude between the source
    and target nodes.

    The resulting direction is represented as a unit vector starting at (0, 0), with X and Y components.

    Attributes
    ----------
    norm : Optional[str]
        Normalization method.
    luse_rotated_features : bool
        Whether to use rotated features.

    Methods
    -------
    compute(graph, source_name, target_name)
        Compute direction of all edges.
    """

    def __init__(self, norm: str | None = None, luse_rotated_features: bool = True) -> None:
        super().__init__(norm)
        self.luse_rotated_features = luse_rotated_features

    def get_raw_values(self, graph: HeteroData, source_name: str, target_name: str) -> np.ndarray:
        """Compute directional features for edges.

        Parameters
        ----------
        graph : HeteroData
            The graph.
        source_name : str
            The name of the source nodes.
        target_name : str
            The name of the target nodes.

        Returns
        -------
        np.ndarray
            The directional features.
        """
        edge_index = graph[(source_name, "to", target_name)].edge_index
        source_coords = graph[source_name].x.numpy()[edge_index[0]].T
        target_coords = graph[target_name].x.numpy()[edge_index[1]].T
        edge_dirs = directional_edge_features(source_coords, target_coords, self.luse_rotated_features).T
        return edge_dirs


class EdgeLength(BaseEdgeAttribute):
    """Edge length feature.

    Attributes
    ----------
    norm : str
        Normalization method.
    invert : bool
        Whether to invert the edge lengths, i.e. 1 - edge_length.

    Methods
    -------
    compute(graph, source_name, target_name)
        Compute edge lengths attributes.
    """

    def __init__(self, norm: str | None = None, invert: bool = False) -> None:
        super().__init__(norm)
        self.invert = invert

    def get_raw_values(self, graph: HeteroData, source_name: str, target_name: str) -> np.ndarray:
        """Compute haversine distance (in kilometers) between nodes connected by edges.

        Parameters
        ----------
        graph : HeteroData
            The graph.
        source_name : str
            The name of the source nodes.
        target_name : str
            The name of the target nodes.

        Returns
        -------
        np.ndarray
            The edge lengths.
        """
        edge_index = graph[(source_name, "to", target_name)].edge_index
        source_coords = graph[source_name].x.numpy()[edge_index[0]]
        target_coords = graph[target_name].x.numpy()[edge_index[1]]
        edge_lengths = haversine_distance(source_coords, target_coords)
        return edge_lengths

    def post_process(self, values: np.ndarray) -> torch.Tensor:
        """Post-process edge lengths."""
        if self.invert:
            values = 1 - values
        return super().post_process(values)
